# -*- coding: utf-8 -*-
import arcpy
from datetime import datetime
import os
from importlib import reload

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

class Create_Standardized_GeoDatabase(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Create Geodatabase Tool"
        self.description = "Creates standard and scratch geodatabases in a directory on the user's local computer." 
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions""" 

        output_dir = arcpy.Parameter(
            name="output_dir",
            displayName="Output Directory Location",
            datatype="DEFolder",
            parameterType="Required",
            direction="Input"
        )  

        params = [
            output_dir
          ]

        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def execute(self, parameters, messages):
        """The source code of the tool."""

        # create output directory if it doesn't already exist    
        output_dir = None
        dir_number_created_today = 1
        gdb_name = "Standardized_BMC_Ped_Infra_Assessment.gdb"
        scratch_geodatabase = "Scratch_BMC_Ped_Infra_Assessment.gdb"

        while output_dir is None:
            sub_dir_name = 'ped_infra_strd_gdb_{0}_{1}'.format(datetime.today().strftime('%Y%m%d'), dir_number_created_today)
            attempt_output_location = os.path.join(parameters[0].valueAsText, sub_dir_name)
            
            if not os.path.isdir(attempt_output_location): 
                
                # create directory with today's date and attempt number 
                os.mkdir(attempt_output_location)

                # set output directory 
                output_dir = attempt_output_location
                
                # create geodatabase 
                arcpy.CreateFileGDB_management(output_dir, gdb_name)
                arcpy.CreateFileGDB_management(output_dir, scratch_geodatabase)

                # Message
                arcpy.AddMessage("Success, created a standardized geodatabase here: {0}".format(os.path.join(output_dir, gdb_name)))
                
                # set work directory 
                arcpy.env.workspace = output_dir

                ##### Domains 
                # Percent 
                percent_domain_name =  "Percent"
                arcpy.CreateDomain_management(
                    in_workspace = gdb_name, 
                    domain_name = percent_domain_name, 
                    domain_description = "Float; Range between 0.0 and 100.0.",
                    field_type = "FLOAT",
                    domain_type = "RANGE"
                    )

                arcpy.SetValueForRangeDomain_management(
                    in_workspace = gdb_name, 
                    domain_name = percent_domain_name, 
                    min_value =0., 
                    max_value = 100.
                    )

                arcpy.AddMessage("Added Domain for percent values to: {0}".format(os.path.join(output_dir, gdb_name)))
                
                # Positive Integers  
                pos_ints_domain_name =  "Positive Integers"          
                arcpy.CreateDomain_management(
                    in_workspace = gdb_name,
                    domain_name = pos_ints_domain_name,
                    domain_description = "Long; Range between 0 and 2,147,483,647.",
                    field_type = "LONG",
                    domain_type = "RANGE"
                    )

                arcpy.SetValueForRangeDomain_management(
                    in_workspace = gdb_name,
                    domain_name = pos_ints_domain_name,
                    min_value =0,
                    max_value = 2147483647
                    )

                arcpy.AddMessage("Added Domain for percent positive integers to: {0}".format(os.path.join(output_dir, gdb_name)))

                # Boolean
                boolean_domain_name =  "Boolean (True or False)"     
                arcpy.CreateDomain_management(
                    in_workspace = gdb_name, 
                    domain_name = boolean_domain_name, 
                    domain_description = "Text; Coded true or false values.",
                    field_type = "TEXT",
                    domain_type = "CODED"
                    )

                booleanDict = {"True":"True", "False": "False"}

                for code in booleanDict:  
                    arcpy.AddCodedValueToDomain_management(gdb_name, boolean_domain_name, code, booleanDict[code])

                arcpy.AddMessage("Added domain for boolean values to: {0}".format(os.path.join(output_dir, gdb_name)))

                # Boolean
                mm_domain_name =  "Transit Stop Level (Major vs. Minor)"     
                arcpy.CreateDomain_management(
                    in_workspace = gdb_name, 
                    domain_name = mm_domain_name, 
                    domain_description = "Text; Coded major or minor values.",
                    field_type = "TEXT",
                    domain_type = "CODED"
                    )

                mmDict = {"Major":"Major", "Minor": "Minor"}

                for code in mmDict:  
                    arcpy.AddCodedValueToDomain_management(gdb_name, mm_domain_name, code, mmDict[code])

                arcpy.AddMessage("Added domain for transit stop levels to: {0}".format(os.path.join(output_dir, gdb_name)))

            else:
                dir_number_created_today += 1

        return 