# -*- coding: utf-8 -*-
import arcpy
import xlrd
import os
import pandas as pd
from collections import Counter
from importlib import reload
reload(arcpy)

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

from python.scripts_utilities.importer_config import *
reload(python.scripts_utilities.importer_config)
from python.scripts_utilities.importer_config import *

print_debug_messages = True

def _parseFieldType(esriType):
    if esriType.lower() in ("text","string","date"):
        return "text"
    elif esriType.lower() in ("short","long","integer","smallinteger"):
        return "integer"
    elif esriType.lower() in ("float","single","double"):
        return "real"
    else:
        return "blob"

def _check_for_same_contents(list1, list2):
    """
    Returns true if lists contain the same values, regardless of order 
    args: 
        list1    -- first list
        list2    -- second list 
    """
    # create a list of unique values after combining two lists 
    for x in set(list1 + list2):

        # check for unique value in each list
        if list1.count(x) != list2.count(x):
            return False
    return True

class Import_Data_To_GeoDatabase(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Import to GeoDatabase Tool"
        self.description = "Adds inforamtion to standardized GeoDatabase" 
        self.canRunInBackground = False
        self.project_spatial_ref = spatial_ref
        self.validation = {
            "correct_geodatabases":   {   
                "valid": False,
                "error_message": "Geodatabase names are not valid. This can result from A) incorrect geodatabase names, B) missing geodatabases, C) too many geodatabases in folder. Please refer to the manual for the required geodatabases."
                },
            "table_is_valid":   {   
                "valid": False,
                "error_message": "The feature class or table you selected is not valid."
                },              
            "expected_vals_present":   {   
                "valid": False,
                "error_message": "The worksheet you selected does not contain all the expected attribute names for the data you want to load."
                },    
            "expected_vals_hv_metadata":   {   
                "valid": False,
                "error_message": "Metadata are not available for all expected attributes in the worksheet you provided."
                },    
            "all_atts_exist_in_ftr":   {   
                "valid": False,
                "error_message": "Expected attributes do not all exist in the data you provided."
                },    
            "p_id_is_unique": {
                "valid": False,
                "error_message": "The 'p_id' column is not unique."
            },
            "all_col_headers_present":   {   
                "valid": False,
                "error_message": "The worksheet does not contain the correct headers."
                },  
            "all_col_headers_present_fc":   {   
                "valid": False,
                "error_message": "The feature class does not contain the correct attributes."
                },   
            "only_two_columns":   {   
                "valid": False,
                "error_message": "The worksheet you provided can have only 2 columns. Please update the workbook."
                },    
            "one_column":   {   
                "valid": False,
                "error_message": "The worksheet you provided has only one column. Two are required."
                },
            "table_exists":{
                "valid": False,
                "error_message": "This table already exists. Delete it or select another layer."
            },
            "neg_values":{
                'valid': False,
                "error_message": "Negative values were found where they should not exist." 
            },
            "correct_cat_values": {
                'valid': False,
                "error_message": "There are unexpected categorical values in the data you've selected." 
            },
            "field_data_type": {
                "valid": False,
                "error_message": "An incorrect field data type was located in your data."
            },
            "correct_geometry_type": {
                "valid": False,
                "error_message": "The feature class you uploaded is the incorrect geometry type."
            }
        }

    def getParameterInfo(self):
        """Define parameter definitions""" 

        gdb_dir = arcpy.Parameter(
            name="gdb_dir",
            displayName="Folder with Standardized and Scratch Geodatabases",
            datatype="DEFolder",
            parameterType="Optional",
            enabled=True,
            direction="Input"
        )  

        layer = arcpy.Parameter(
            name="layer",
            displayName="Standard Layer",
            datatype="GPString",
            parameterType="Optional",
            enabled=True,
            direction="Input"
        ) 
        layer.filter.type = "ValueList"
        layer.filter.list = list(layerInfo.keys())

        fc_to_import = arcpy.Parameter(
            name="fc_to_import",
            displayName="Select Feature Class to Import",
            datatype="GPFeatureLayer",
            parameterType="Optional",
            enabled=False,
            direction="Input"
        ) 

        excel_workbook = arcpy.Parameter(
            name="excel_workbook",
            displayName="Excel Workbook",
            datatype="DEFile",
            parameterType="Optional",
            enabled=True,
            direction="Input"
        ) 
        excel_workbook.filter.list = ['xlsx']

        worksheet = arcpy.Parameter(
            name="worksheet",
            displayName="Worksheet with Location Information",
            datatype="GPString",
            parameterType="Optional",
            enabled=True,
            direction="Input"
        ) 
        worksheet.filter.type = "ValueList"
        worksheet.filter.list = ["Select Excel Worksheet Containing with the Data"]

        params = [
            gdb_dir,
            layer,
            fc_to_import,
            excel_workbook,
            worksheet
          ]

        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""

        # show excel parameters for tables 
        if parameters[1].valueAsText in getKeysFromValue(layerInfo, "processType", "excelTable") and parameters[1].valueAsText != None:
            parameters[2].enabled = False
            parameters[3].enabled = True
            parameters[4].enabled = True
        else: 
            parameters[2].enabled = True
            parameters[3].enabled = False
            parameters[4].enabled = False

        # check extention type of excel workbook, must be .xlwx
        if parameters[3].valueAsText != None and parameters[3].enabled: 
            excelbook_info = arcpy.Describe(parameters[3].valueAsText)
            if excelbook_info.extension in ["xlsx"]:
                workbook = xlrd.open_workbook(parameters[3].valueAsText)
                parameters[4].filter.list = list(workbook.sheet_names())
                
            else:
                arcpy.AddError("Workbook must be an Excel workbook with the file extension type, '.xslx'")
            
    def execute(self, parameters, messages):
        """The source code of the tool."""

        # create list of geodatabases 
        dict_of_expected_geodatabases = {
            "scratch_geodatabase": os.path.join(parameters[0].valueAsText, "Scratch_BMC_Ped_Infra_Assessment.gdb"),
            "final_geodatabase": os.path.join(parameters[0].valueAsText, "Standardized_BMC_Ped_Infra_Assessment.gdb")
        }

        # set descriptions
        gdb_info = arcpy.Describe(parameters[0].value)

        # Set environment settings
        arcpy.env.workspace = parameters[0].valueAsText 
        workspaces = arcpy.ListWorkspaces("*", "FileGDB")
        arcpy.env.overwriteOutput = True

        # ensure exected geodatabases exists
        if collections.Counter(workspaces) == collections.Counter(list(dict_of_expected_geodatabases.values())):
            self.validation["correct_geodatabases"]["valid"] = True
        else:
            arcpy.AddError(self.validation["correct_geodatabases"]["error_message"])
            return

        # upload the feature classes if only geometries are needed (no transformation)
        if parameters[1].value in getKeysFromValue(layerInfo, "processType", "NeedOnlyGeoms") and self.validation["correct_geodatabases"]["valid"]: 
            arcpy.env.workspace = dict_of_expected_geodatabases["final_geodatabase"]
            arcpy.env.overwriteOutput = False
            fc_to_import = arcpy.Describe(parameters[2].value)
            expected_geometry = getExpectedGeometry(layerInfo, parameters[1].valueAsText)

            # ensure the table doesn't already exist
            if print_debug_messages:
                arcpy.AddMessage(f"The layer process type is 'NeedOnlyGeoms'")
                arcpy.AddMessage(f"Current workspace is '{arcpy.env.workspace}'")
            if arcpy.Exists(os.path.join(dict_of_expected_geodatabases["final_geodatabase"], getLayerName(layerInfo, parameters[1].valueAsText))):  
                arcpy.AddError(self.validation["table_exists"]["error_message"])
                return
            else:
                self.validation["table_exists"]["valid"] = True  

            # ensure the feature class is non empty
            if int(arcpy.management.GetCount(parameters[2].value)[0]) > 0: 
                self.validation["table_is_valid"]["valid"] = True  
            else:
                arcpy.AddError(self.validation["table_is_valid"]["error_message"])
                return

            # ensure geometry type is correct
            if fc_to_import.ShapeType in expected_geometry:
                self.validation["correct_geometry_type"]["valid"] = True
            else:
                arcpy.AddError(self.validation["correct_geometry_type"]["error_message"] + " The data type of your data is: {0}, it should be {1}.".format(fc_to_import.ShapeType, expected_geometry))  
                return

            arcpy.env.workspace = dict_of_expected_geodatabases["scratch_geodatabase"]
            arcpy.env.overwriteOutput = True

            # load data into geodatabase  
            arcpy.conversion.FeatureClassToGeodatabase(
                Input_Features = parameters[2].valueAsText, 
                Output_Geodatabase = dict_of_expected_geodatabases["scratch_geodatabase"]
                )

            # reproject data 
            arcpy.management.Project(
                in_dataset = parameters[2].valueAsText,  
                out_dataset = "{0}_{1}".format(getLayerName(layerInfo, parameters[1].valueAsText), 'projected'), 
                out_coor_system = self.project_spatial_ref,
                preserve_shape = "PRESERVE_SHAPE",
                vertical = "NO_VERTICAL"
                )

            # send to final resting place 
            arcpy.conversion.FeatureClassToGeodatabase(
                Input_Features = "{0}_{1}".format(getLayerName(layerInfo, parameters[1].valueAsText), 'projected'), 
                Output_Geodatabase = dict_of_expected_geodatabases["final_geodatabase"]
                )

            arcpy.env.workspace = dict_of_expected_geodatabases["final_geodatabase"]
            arcpy.env.overwriteOutput = False

            # rename the file             
            arcpy.management.Rename(
                in_data =  "{0}_{1}".format(getLayerName(layerInfo, parameters[1].valueAsText), 'projected'), 
                out_data =  getLayerName(layerInfo, parameters[1].valueAsText), 
                data_type = "FeatureClass"
                )

            arcpy.management.RepairGeometry(
                in_features =  getLayerName(layerInfo, parameters[1].valueAsText),
                delete_null = "DELETE_NULL",
                validation_method = "ESRI"
                )

            arcpy.AddMessage("Successfully added the {0} table to {1}.gdb as {2} layer.".format(fc_to_import.baseName, gdb_info.baseName, parameters[1].valueAsText))

        # else if inputs are tabular data, load them 
        elif parameters[1].valueAsText in getKeysFromValue(layerInfo, "processType", "excelTable") :
            
            # set workspace enviornment 
            arcpy.env.workspace = dict_of_expected_geodatabases["final_geodatabase"]
            arcpy.env.overwriteOutput = False
            excelbook_info = arcpy.Describe(parameters[3].value)
            selected_target_sheet = parameters[4].value
            expected_atts = getRequiredAttributes(layerInfo, parameters[1].valueAsText)
            # open spreadsheet
            df = pd.DataFrame()
            df = pd.read_excel(
                io = parameters[3].valueAsText, 
                sheet_name = selected_target_sheet,
                header=0
                )

            # ensure the table doesn't already exist
            if print_debug_messages:
                arcpy.AddMessage(f"The layer process type is 'excelTable'")
                arcpy.AddMessage(f"Current workspace is '{arcpy.env.workspace}'")
            if arcpy.Exists(os.path.join(dict_of_expected_geodatabases["final_geodatabase"], getLayerName(layerInfo, parameters[1].valueAsText))): 
                arcpy.AddError(self.validation["table_exists"]["error_message"])
                return
            else:
                self.validation["table_exists"]["valid"] = True  

            # ensure the worksheet is not empty and only 2 columns are present
            if len(df.columns) != 2: 
                arcpy.AddError(self.validation["only_two_columns"]["error_message"])
                return
            else:
                self.validation["only_two_columns"]["valid"] = True       

            # check to make sure the expected attributes exist in the spreadsheet
            if collections.Counter(list(df.columns)) == collections.Counter(getRequiredAttributes(layerInfo, parameters[1].valueAsText)):
                self.validation["all_col_headers_present"]["valid"] = True
            else:
                arcpy.AddError(self.validation["all_col_headers_present"]["error_message"] + ". The table should include the following column headers: {}. The worksheet you identified contains: {}".format(getRequiredAttributes(layerInfo, parameters[1].valueAsText), list(df.columns)))  
                return       

            # check p key is unique in worksheet
            if pd.Series(df["p_id"]).is_unique:
                self.validation["p_id_is_unique"]["valid"] = True
            else:
                arcpy.AddError(self.validation["p_id_is_unique"]["error_message"])  
                return       

            # check to make sure the expected values exist in each column
            for i in getRequiredAttributes(layerInfo, parameters[1].valueAsText):
                ats = getExpectedAttrValues(layerInfo, parameters[1].valueAsText, i)
                if ats is not None:
                    if all(x in list(set(ats)) for x in list(set(df.iloc[:, 1]))):
                    # if _check_for_same_contents(list(set(df.iloc[:, 1])), list(set(ats))):
                        self.validation["all_atts_exist_in_ftr"]["valid"] = True
                    elif df.iloc[:, 1].isnull().values.any():
                        arcpy.AddError(self.validation["all_atts_exist_in_ftr"]["error_message"] + ' Attribute values should be: {0} and your data seems to have empty cells. Please correct them to match the expected values.'.format(ats)) 
                        return
                    else:
                        arcpy.AddError(self.validation["all_atts_exist_in_ftr"]["error_message"] + ' Attribute values should be: {0} and your data values are: {1}. Please correct them to match the expected values.'.format(ats, list(set(df.iloc[:, 1]))))  
                        return

            # check to make sure id column is unique 
            if is_pd_field_unqiue(df, 'p_id'):
                self.validation["p_id_is_unique"]["valid"] = True
            else:
                arcpy.AddError(self.validation["p_id_is_unique"]["error_message"]) 
                return      

            arcpy.ExcelToTable_conversion(
                Input_Excel_File = parameters[3].valueAsText, 
                Output_Table =  getLayerName(layerInfo, parameters[1].valueAsText), 
                Sheet = parameters[4].valueAsText
                )

            arcpy.AddMessage("Successfully added the {0} worksheet from the {1}.xlsx workbook to {2}.gdb as the {3} layer.".format(parameters[4].valueAsText, excelbook_info.baseName, gdb_info.baseName, parameters[1].valueAsText))


        else:
            arcpy.env.workspace = dict_of_expected_geodatabases["final_geodatabase"]
            arcpy.env.overwriteOutput = False
            fc_to_import = arcpy.Describe(parameters[2].value)
            attributes = getRequiredAttributes(layerInfo, parameters[1].valueAsText)
            fc_attributes =[f.name for f in arcpy.ListFields(parameters[2].valueAsText)]  
            expected_geometry = getExpectedGeometry(layerInfo, parameters[1].valueAsText)
            
            # ensure the table doesn't already exist
            if print_debug_messages:
                arcpy.AddMessage(f"The layer process type is 'CheckAttributes'")
                arcpy.AddMessage(f"Current workspace is '{arcpy.env.workspace}'")
            if arcpy.Exists(os.path.join(dict_of_expected_geodatabases["final_geodatabase"], getLayerName(layerInfo, parameters[1].valueAsText))): 
                arcpy.AddError(self.validation["table_exists"]["error_message"])
                return
            else:
                self.validation["table_exists"]["valid"] = True  

            # ensure file is not empty
            if int(arcpy.management.GetCount(parameters[2].value)[0]) > 0: 
                self.validation["table_is_valid"]["valid"] = True
            else:
                arcpy.AddError(self.validation["table_is_valid"]["error_message"])
                return

            # ensure expected attributes exist
            if all(w in fc_attributes for w in attributes):
                self.validation["all_col_headers_present_fc"]["valid"] = True
            else:
                arcpy.AddError(self.validation["all_col_headers_present_fc"]["error_message"] + ' You data you attempted to upload contains the following headers: {0}. It should contain at least the following: {1}.'.format(fc_attributes, attributes))  
                return

            # ensure geometry type is correct
            if fc_to_import.ShapeType in expected_geometry:
                self.validation["correct_geometry_type"]["valid"] = True
            else:
                arcpy.AddError(self.validation["correct_geometry_type"]["error_message"])  
                return

            # ensure primary key contains only unique values
            # if is_field_unique(parameters[2].valueAsText, 'p_id'):
            #     self.validation["p_id_is_unique"]["valid"] = True
            # else:
            #     arcpy.AddError(self.validation["p_id_is_unique"]["error_message"]) 
            #     return   

            # ensure data types are correct 
            # for each attribute of the input layer, ensure incoming datatype is allowed
            df = feature_class_to_df(parameters[2].valueAsText, attributes, 0)
            for i, v in enumerate(attributes):
                index = [i.name for i in arcpy.ListFields(parameters[2].valueAsText)].index(v)
                current_attribute_type = _parseFieldType(arcpy.ListFields(parameters[2].valueAsText)[index].type)
                expect_data_type = getExpectedAttrDataType(layerInfo, parameters[1].valueAsText, v)

                if expect_data_type is not None:
                    if current_attribute_type in expect_data_type and i + 1 != len(attributes): # plus one becuase length of empty list is 0. We have already confirmed attributes exists by this point
                        pass # move on 
                    elif current_attribute_type in expect_data_type and i + 1 == len(attributes): # if the attirbutes match at we're at the end of the list
                        self.validation["field_data_type"]["valid"] = True
                    else:
                        arcpy.AddError(self.validation["field_data_type"]["error_message"] + '. The data type of the {0} attribute should be: {1}. The data type is currently, {2}.'.format(v, expect_data_type, current_attribute_type))
                        return 
                elif i + 1 == len(attributes): # if there's no expected attribute and we're at the end of the list 
                    self.validation["field_data_type"]["valid"] = True

            # ensure unexpected values are not present
            for i, v in enumerate(attributes):
                index = [ind.name for ind in arcpy.ListFields(parameters[2].valueAsText)].index(v)
                # ft = arcpy.ListFields(parameters[2].valueAsText)[index].type 
                ex = getExpectedAttrValues(layerInfo, parameters[1].valueAsText, v)
                if ex == 'Non Negative':
                    if df[v].any() < 0: 
                        arcpy.AddError(self.validation["neg_values"]["error_message"] + '. Negative value found in attribute, {0}.'.format(v))
                        return 
                    elif i + 1 == len(attributes):
                        self.validation["neg_values"]["valid"] = True
                elif ex == 'Categorical':
                    if len(list(set(df[i])) - set(ex)) > 0: 
                        arcpy.AddError(self.validation["all_atts_exist_in_ftr"]["error_message"])
                        return      
                    elif i + 1 == len(attributes):
                        self.validation["correct_cat_values"]["valid"] = True
                elif i + 1 == len(attributes):
                    self.validation["correct_cat_values"]["valid"] = True
 
            arcpy.AddMessage("Validation complete. Adding feature class to geodatabase.")

            # # todo add metadata to feature class 
            arcpy.env.workspace = dict_of_expected_geodatabases["scratch_geodatabase"]
            arcpy.env.overwriteOutput = True

            # load data into geodatabase  
            arcpy.conversion.FeatureClassToGeodatabase(
                Input_Features = parameters[2].valueAsText, 
                Output_Geodatabase = dict_of_expected_geodatabases["scratch_geodatabase"]
                )

            # reproject data 
            arcpy.management.Project(
                in_dataset = parameters[2].valueAsText,  
                out_dataset = "{0}_{1}".format( getLayerName(layerInfo, parameters[1].valueAsText), 'projected'), 
                out_coor_system = self.project_spatial_ref,
                preserve_shape = "PRESERVE_SHAPE",
                vertical = "NO_VERTICAL"
                )

            # send to final resting place 
            arcpy.conversion.FeatureClassToGeodatabase(
                Input_Features = "{0}_{1}".format( getLayerName(layerInfo, parameters[1].valueAsText), 'projected'), 
                Output_Geodatabase = dict_of_expected_geodatabases["final_geodatabase"]
                )
 
            arcpy.env.workspace = dict_of_expected_geodatabases["final_geodatabase"]
            arcpy.env.overwriteOutput = True

            # rename the file             
            arcpy.management.Rename(
                in_data =  "{0}_{1}".format( getLayerName(layerInfo, parameters[1].valueAsText), 'projected'), 
                out_data =  getLayerName(layerInfo, parameters[1].valueAsText), 
                data_type = "FeatureClass"
                )

            arcpy.management.RepairGeometry(
                in_features =  getLayerName(layerInfo, parameters[1].valueAsText),
                delete_null = "DELETE_NULL",
                validation_method = "ESRI"
                )

            arcpy.AddMessage("Successfully added the {0} feature class to {1}.gdb.".format(fc_to_import.baseName, gdb_info.baseName))

        return 

