# -*- coding: utf-8 -*-
import arcpy
import os
import shutil
import pandas as pd
from openpyxl import load_workbook
from importlib import reload

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

import python.scripts_utilities.importer_config
reload(python.scripts_utilities.importer_config)
from python.scripts_utilities.importer_config import *

import python.scripts_utilities.prioritization_config
reload(python.scripts_utilities.prioritization_config)
from python.scripts_utilities.prioritization_config import * 


class Prioritization_Create_Workbook(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Create Prioritization Workbook Tool"
        self.description = "Returns an Excel file with spatial processing complete for prioritization."
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""

        # 1 Features to prioritize
        prioritizion_features = arcpy.Parameter(
            name="prioritizion_features",
            displayName="Prioritization Feature Layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input"
        )    

        primary_key = arcpy.Parameter(
            name="primary_key",
            displayName="Primary Key",
            datatype="GPString",
            parameterType="Required",
            direction="Input"
        ) 

        primary_key.filter.type = "ValueList"
        primary_key.filter.list = ["Select Layer of Locations to Prioritize"]

        gdb_dir = arcpy.Parameter(
            name="gdb_dir",
            displayName="Folder with Standardized and Scratch Geodatabases",
            datatype="DEFolder",
            parameterType="Required",
            enabled=True,
            direction="Input"
        )

        # excel_template = arcpy.Parameter(
        #     name="excel_template",
        #     displayName="Excel Workbook Template",
        #     datatype="DEFile",
        #     parameterType="Required",
        #     enabled=True,
        #     direction="Input"
        # ) 
        # excel_template.filter.list = ['xlsx']

        output_excel = arcpy.Parameter(
            name="output_excel",
            displayName="Output Prioritization Workbook",
            datatype="DEFile",
            parameterType="Required",
            enabled=True,
            direction="Output"
        ) 
        output_excel.filter.list = ['xlsx']
        # output_excel.filter.list = ['xlsm']

        # Opportunity
        factorOpportunity = arcpy.Parameter(
            name="factorOpportunity",
            displayName="FACTOR - Whether to include opportunities factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Opportunities"
        )   

        # Opportunity - Grant Funding Measure
        measureGrantFunding = arcpy.Parameter(
            name="measureGrantFunding",
            displayName="   VARIABLE - Whether to include grant funding variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Opportunities"
        )   

        # Opportunity - Coincides with Planned Roadway Improvement / Maintenance 
        measureCoincidesRoadwayImprovement = arcpy.Parameter(
            name="measureCoincidesRoadwayImprovement",
            displayName="   VARIABLE - Whether to include roadway improvement variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Opportunities"
        )    

        # Constraints
        factorConstraints = arcpy.Parameter(
            name="factorConstraints",
            displayName="FACTOR - Whether to include constraints factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Constraints"
        )  

        # Constraints - Advisory Committee Measure
        measureMultiJurisdictional = arcpy.Parameter(
            name="measureMultiJurisdictional",
            displayName="   VARIABLE - Whether to include multi jurisditional coordination variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Constraints"
        )   

        # Constraints - Advisory Committee Measure
        measureCost = arcpy.Parameter(
            name="measureCost",
            displayName="   VARIABLE - Whether to include cost variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Constraints"
        )   

        # Stakeholder 
        factorStakeholder = arcpy.Parameter(
            name="factorStakeholder",
            displayName="FACTOR - Whether to include stakeholder factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Stakeholder Input"
        ) 

        # Stakeholder - Advisory Committee Measure
        measureAdvisoryCommittee = arcpy.Parameter(
            name="measureAdvisoryCommittee",
            displayName="   VARIABLE - Whether to include advisory committee recommendation variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Stakeholder Input"
        )   

        # Stakeholder - Adopted Plan 
        measureAdoptedPlan = arcpy.Parameter(
            name="measureAdoptedPlan",
            displayName="   VARIABLE - Whether to include adopted plan variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Stakeholder Input"
        )   

        ## Equity 
        factorEquity = arcpy.Parameter(
            name="factorEquity",
            displayName="FACTOR - Whether to include equity factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        )  

        # Equity - People with Disabilities
        measureDisabilities = arcpy.Parameter(
            name="measureDisabilities",
            displayName="   VARIABLE - Whether to include people with disabilities variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        )

        # Equity - Households Below Poverty
        measureIncome = arcpy.Parameter(
            name="measureIncome",
            displayName="   VARIABLE - Whether to include households below poverty variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        )

        # Equity - Limited English Proficiency
        measureLimitedEnglishProficiency = arcpy.Parameter(
            name="measureLimitedEnglishProficiency",
            displayName="   VARIABLE - Whether to include limited english proficiency variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        )

        # Equity - Limited English Proficiency
        measureVehicleAccess = arcpy.Parameter(
            name="measureVehicleAccess",
            displayName="   VARIABLE - Whether to include zero car households variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        )

        # Equity - over 75 years
        measureOlderPopulation = arcpy.Parameter(
            name="measureOlderPopulation",
            displayName="   VARIABLE - Whether to include population over 75 years variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        )

        # Equity - Hispanic
        measureHispanicPopulation = arcpy.Parameter(
            name="measureHispanicPopulation",
            displayName="   VARIABLE - Whether to include Hispanic population variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        ) 

        # Equity - Non-white
        measureNonWhitePopulation = arcpy.Parameter(
            name="measureNonWhitePopulation",
            displayName="   VARIABLE - Whether to include non-white population variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        ) 

        # Equity - Under 18 population 
        measureYoungerPopulation = arcpy.Parameter(
            name="measureYoungerPopulation",
            displayName="   VARIABLE - Whether to include under 18 population variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        ) 

        # Equity - Vulnerable Population Index
        measureVPI = arcpy.Parameter(
            name="measureVPI",
            displayName="   VARIABLE - Whether to include Vulnerable Population Index variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Equity"
        ) 

        ## Connectivity 
        factorConnectivity = arcpy.Parameter(
            name="factorConnectivity",
            displayName="FACTOR - Whether to include connectivity factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Connectivity"
        )  

        # Connectivity - Connects to Existing Facility
        measurePedestrianNetwork = arcpy.Parameter(
            name="measurePedestrianNetwork",
            displayName="   VARIABLE - Whether to include connections to pedestrian network as variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Connectivity"
        )  

        # Connectivity - Street Connectivity (Intersection Density)
        measureIntersectionDensity = arcpy.Parameter(
            name="measureIntersectionDensity",
            displayName="   VARIABLE - Whether to include intersection density as variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Connectivity"
        )  

        # Pedestrian Trip Potential 
        factorPedTripPotential = arcpy.Parameter(
            name="factorPedTripPotential",
            displayName="FACTOR - Whether to include pedestrian trip potential factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )   

        # Pedestrian Trip Potential - Population Density
        measurePopulationDensity = arcpy.Parameter(
            name="measurePopulationDensity",
            displayName="   VARIABLE - Whether to include population density variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Pedestrian Trip Potential - Employment Density
        measureEmploymentDensity = arcpy.Parameter(
            name="measureEmploymentDensity",
            displayName="   VARIABLE - Whether to include employment density variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Pedestrian Trip Potential - K - 12 Schools
        measureSchools = arcpy.Parameter(
            name="measureSchools",
            displayName="   VARIABLE - Whether to include K - 12 schools variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Pedestrian Trip Potential - Proximity to Colleges and Universities
        measureHigherEducation = arcpy.Parameter(
            name="measureHigherEducation",
            displayName="   VARIABLE - Whether to include colleges and universities variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Pedestrian Trip Potential - Proximity to Retail
        measureRetail = arcpy.Parameter(
            name="measureRetail",
            displayName="   VARIABLE - Whether to include proximity to retail variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Pedestrian Trip Potential - Minor Transit Stops
        measureMinorTransit = arcpy.Parameter(
            name="measureMinorTransit",
            displayName="   VARIABLE - Whether to include prxomity to minor transit stops variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Pedestrian Trip Potential - Proximity to Colleges and Universities
        measureMajorTransit = arcpy.Parameter(
            name="measureMajorTransit",
            displayName="   VARIABLE - Proximity to major transit stops variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Pedestrian Trip Potential"
        )  

        # Safety
        factorSafety = arcpy.Parameter(
            name="factorSafety",
            displayName="FACTOR - Whether to include safety factor in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Safety"
        ) 

        # Safety - Total Pedestrian Crashes
        measureCrashes = arcpy.Parameter(
            name="measureCrashes",
            displayName="   VARIABLE - Whether to include pedestrian crash variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Safety"
        ) 

        # Safety - Total Fatal and Severe Pedestrian Crashes
        measureSevereFatalCrashes = arcpy.Parameter(
            name="measureSevereFatalCrashes",
            displayName="   VARIABLE - Whether to include fatal and severe pedestrian crash variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Safety"
        ) 

        # Safety - Functional Classification
        measureFunctionalClass = arcpy.Parameter(
            name="measureFunctionalClass",
            displayName="   VARIABLE - Whether to include functional class variable in prioritization",
            datatype="GPBoolean",
            parameterType="Optional",
            enabled=True,
            direction="Input",
            category="Safety"
        )

        params = [
            prioritizion_features,
            primary_key,
            gdb_dir,
            # excel_template,
            output_excel,
            factorOpportunity,
            measureGrantFunding,
            measureCoincidesRoadwayImprovement,
            factorConstraints,
            measureMultiJurisdictional,
            measureCost,
            factorStakeholder,
            measureAdvisoryCommittee,
            measureAdoptedPlan,
            factorEquity,
            measureDisabilities,
            measureIncome,
            measureLimitedEnglishProficiency,
            measureVehicleAccess,
            measureOlderPopulation,
            measureHispanicPopulation,
            measureNonWhitePopulation,
            measureYoungerPopulation,
            measureVPI,
            factorConnectivity,
            measurePedestrianNetwork,
            measureIntersectionDensity,
            factorPedTripPotential,
            measurePopulationDensity,
            measureEmploymentDensity,
            measureSchools,
            measureHigherEducation,
            measureRetail,
            measureMinorTransit,
            measureMajorTransit,
            factorSafety,
            measureCrashes,
            measureSevereFatalCrashes,
            measureFunctionalClass,
            ]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""


        if parameters[0].valueAsText != None and parameters[0].enabled: 
            fc_attributes =[f.name for f in arcpy.ListFields(parameters[0].valueAsText) if not f.required] 
            parameters[1].filter.list = fc_attributes
        
        # TODO consider commenting this out if we want the user to explicitly give the location of output file
        # if parameters[2].value and parameters[3].value is None:
        #     parameters[3].value = os.path.join(parameters[2].valueAstext, "Prioritization_Results.xlsx")

        for i in sub_params.keys():
            if _get_by_name(parameters, i).value:
                for j in sub_params[i]:
                    _get_by_name(parameters, j).enabled = True
            else:
                for j in sub_params[i]:
                    _get_by_name(parameters, j).enabled = False
                    _get_by_name(parameters, j).value = False

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        # list = [parameters[i].name for i in range(len(parameters))]
        prioritizion_features = _get_by_name(parameters, "prioritizion_features").valueAsText
        primary_key = _get_by_name(parameters, "primary_key").valueAsText
        gdb_dir = _get_by_name(parameters, "gdb_dir").valueAsText
        output_excel = _get_by_name(parameters, "output_excel").valueAsText

        excel_template = os.path.join(template_path, "BMC_Prioritization_Template_2016.xlsx")
        # # make a copy of the template to write data into
        # shutil.copy(excel_template, os.path.join(gdb_dir, "tmp_workbook.xlsx"))

        wb = load_workbook(excel_template)
        ws = wb["Inputs - Weights"]
        # fill in table for variables that are to be included
        for i in sub_params.keys():
            for j, v in sub_params[i].items():
                if _get_by_name(parameters, j).value:
                    ws[f"{variable_include_col}{v['cell_index']}"] = True
                else:
                    ws[f"{variable_include_col}{v['cell_index']}"] = False

        # ## ensure excel template is set up right
        # template_wb = load_workbook(excel_template, read_only=True)
        # # check it has the correct worksheets
        # inputs_sheet = "Inputs - Weights",
        # raw_scores_sheet = "Raw_Scores"
        # ws_list = [inputs_sheet, raw_scores_sheet, "WeightedScores - Variables", "WeightedScores - Factors"]
        # if len([s for s in ws_list if s not in template_wb.sheetnames]) > 0:
        #     arcpy.AddError(f"Excel Template provided does not have a sheet with titles - '{[s for s in ws_list if s not in template_wb.sheetnames]}'.\n It has the following sheets {template_wb.sheetnames}")
        # # check that any factors selected have weights in template
        # input_wt_ws = template_wb[inputs_sheet]
        # for i in sub_params.keys():
        #     if _get_by_name(parameters, i).value and (input_wt_ws[f"{factor_wt_col}{factor_params[i]['wt_cell_row']}"].value > 0):



        # ensure the proper gdb exists in the location specified
        standard_gdb = os.path.join(gdb_dir, "Standardized_BMC_Ped_Infra_Assessment.gdb")
        scratch_gdb = os.path.join(gdb_dir, "Scratch_BMC_Ped_Infra_Assessment.gdb")

        # Set environment settings
        arcpy.env.workspace = gdb_dir
        workspaces = arcpy.ListWorkspaces("*", "FileGDB")
        arcpy.env.overwriteOutput = True

        if not bool(set(workspaces) == set([standard_gdb, scratch_gdb])):
            arcpy.AddError("Geodatabase names are not valid. This can result from A) incorrect geodatabase names, B) missing geodatabases, C) too many geodatabases in folder. Please refer to the manual for the required geodatabases.")
            raise arcpy.ExecuteError

        shp_info = arcpy.Describe(prioritizion_features)
        if shp_info.ShapeType in ["Point", "Multipoint"]:
            shp_type = "Points"
        elif shp_info.ShapeType in ["Polyline", "Multipolyline"]:
            shp_type = "Lines"
        else:
            arcpy.AddError('The shape type of the shapefile you uploaded is not one of the following formats: {0}'.format(["Polyline", "Multipolyline", "Point", "Multipoint"]))
            raise arcpy.ExecuteError
        
        # check if field is unique
        # TODO check if this is necessary?
        if not is_field_unique(fc=prioritizion_features, field=primary_key):
            arcpy.AddError("The primary key field supplied is either missing value or is being used by more than one features.")
            raise arcpy.ExecuteError
        
        # convert feature class to a pandas array
        # df = feature_class_to_df(in_fc = parameters[0].valueAsText, fields=[primary_key], nulls=-9999)
        # arcpy.AddMessage(df.head())

        # create a copy of the prioritization layer in the scratch database
        tmp_prioritization_layer = os.path.join(scratch_gdb, "tmp_prioritization_layer")
        copy_fc_with_fields(prioritizion_features, tmp_prioritization_layer, fields=[primary_key], scratch_gdb=scratch_gdb)

        # check if there is a need doing a pairwise intersect so that this is not repeated multiple times later
        if _get_by_name(parameters, "factorEquity").value:
            eq_fields_dict = {}
            for k, v in sub_params["factorEquity"].items():
                if _get_by_name(parameters, k).value:
                    eq_fields_dict[k] = v["column"]
                    vpi_table = v["table"]

            if len(eq_fields_dict) > 0:
                tmp_equity_buffer_all_fields = os.path.join(scratch_gdb, "tmp_equity_buffer_all_fields")
                arcpy.AddFields_management(tmp_prioritization_layer, [[f"{f}", "DOUBLE"] for f in eq_fields_dict.keys()])
                arcpy.Buffer_analysis(tmp_prioritization_layer, tmp_equity_buffer_all_fields, "0.5 miles")
                tmp_equity_buffer = os.path.join(scratch_gdb, "tmp_equity_buffer")
                copy_fc_with_fields(in_fc=tmp_equity_buffer_all_fields, out_fc=tmp_equity_buffer, fields=[primary_key]+[f"{f}" for f in eq_fields_dict.keys()], scratch_gdb=scratch_gdb)
                # create a copy of VPI layer with necessary density fields
                tmp_vpi_layer = os.path.join(scratch_gdb, "tmp_vpi_layer")
                vpi_layer = os.path.join(standard_gdb, vpi_table)
                if not arcpy.Exists(vpi_layer):
                    arcpy.AddError(f"Data corresponding to '[{', '.join(eq_fields_dict.keys())}]' is not present in standard GDB. \n Expected dataset: '{vpi_layer}'")
                    raise arcpy.ExecuteError
                copy_fc_with_fields(in_fc=vpi_layer, out_fc=tmp_vpi_layer, fields=list(eq_fields_dict.values())+["p_id"], scratch_gdb=scratch_gdb)

                calculate_density_by_intersect(
                    input_fc=tmp_equity_buffer,
                    input_uid=primary_key,
                    target_fc=tmp_vpi_layer,
                    target_uid="p_id",
                    field_dict=eq_fields_dict,
                    scratch_gdb=scratch_gdb,
                    area_multiplier=43560   # this is to get per acre value instead of per sq ft
                )
            
        # check if there is a need for a half mile buffer and create the layer here to avoid repeating the process here
        half_mile_buffer_needs = [
            "measureIntersectionDensity",
            "measurePopulationDensity",
            "measureEmploymentDensity"
        ]
        if sum([int(_get_by_name(parameters, p).value) for p in half_mile_buffer_needs]):
            tmp_half_mile_buffer = os.path.join(scratch_gdb, "tmp_half_mile_buffer")
            arcpy.Buffer_analysis(tmp_prioritization_layer, tmp_half_mile_buffer, "0.5 miles")
        
                # for f in eq_fields:
                #     arcpy.CalculateField_management(tmp_vpi_layer, f"{f}_per_acre", f"!{f}!/!shape.area@acres!", "PYTHON3", field_type="DOUBLE")
                # arcpy.DeleteField_management(tmp_vpi_layer, eq_fields)

                # # do pariwaise intersect and 
                # tmp_vpi_pw_intersect = os.path.join(scratch_gdb, "tmp_vpi_pw_intersect")


        # dictionary to convert variables into 1 or 0 to deal with any data format inconsistencies
        bool_to_int = {
            "true" : 1,
            "TRUE" : 1,
            "True" : 1,
            "false" : 1,
            "FALSE" : 1,
            "False" : 1,
            1 : 1,
            0 : 0,
            True : 1,
            False : 0
        }

        for i in sub_params.keys():
            for j in sub_params[i].keys():
                if _get_by_name(parameters, j).value:
                    if  j == 'measureGrantFunding':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # join table, mark true at places with matching IDs
                        out_field = j
                        arcpy.AddField_management(tmp_prioritization_layer, field_name=out_field, field_type="SHORT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        target_pkey, target_field = getRequiredAttributes(layerInfo, sub_params[i][j]['name'])
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=out_field,
                            to_pkey=primary_key,
                            from_table=target_table,
                            from_field=target_field,
                            from_pkey=target_pkey,
                            transform_dict=bool_to_int
                        )

                    if j == 'measureCoincidesRoadwayImprovement':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # join table, mark true at places with matching IDs
                        out_field = j
                        arcpy.AddField_management(tmp_prioritization_layer, field_name=out_field, field_type="SHORT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        target_pkey, target_field = getRequiredAttributes(layerInfo, sub_params[i][j]['name'])
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=out_field,
                            to_pkey=primary_key,
                            from_table=target_table,
                            from_field=target_field,
                            from_pkey=target_pkey,
                            transform_dict=bool_to_int
                        )

                    if  j == 'measureMultiJurisdictional':
                        #TODO add a warning for this if it prioritization layer is points type
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # join table, mark true at places with matching IDs
                        out_field = j
                        arcpy.AddField_management(tmp_prioritization_layer, field_name=out_field, field_type="SHORT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=out_field,
                            input_uid=primary_key,
                            target_fc=target_table,
                            agg="sum",
                            scratch_gdb=scratch_gdb
                        )
                        arcpy.CalculateField_management(tmp_prioritization_layer, out_field, f"bool(!{out_field}! <= 1)", "PYTHON3")
                        
                    if j == 'measureCost':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # join table, join cost value at places with matching IDs
                        out_field = j
                        arcpy.AddField_management(tmp_prioritization_layer, field_name=out_field, field_type="TEXT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        target_pkey, target_field = getRequiredAttributes(layerInfo, sub_params[i][j]['name'])
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=out_field,
                            to_pkey=primary_key,
                            from_table=target_table,
                            from_field=target_field,
                            from_pkey=target_pkey
                        )

                    if  j == 'measureAdvisoryCommittee':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # join table, mark true at places with matching IDs
                        out_field = j
                        arcpy.AddField_management(tmp_prioritization_layer, field_name=out_field, field_type="SHORT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        target_pkey, target_field = getRequiredAttributes(layerInfo, sub_params[i][j]['name'])
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=out_field,
                            to_pkey=primary_key,
                            from_table=target_table,
                            from_field=target_field,
                            from_pkey=target_pkey,
                            transform_dict=bool_to_int
                        )

                    if j == 'measureAdoptedPlan':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # join table, mark true at places with matching IDs
                        out_field = j
                        arcpy.AddField_management(tmp_prioritization_layer, field_name=out_field, field_type="SHORT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        target_pkey, target_field = getRequiredAttributes(layerInfo, sub_params[i][j]['name'])
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=out_field,
                            to_pkey=primary_key,
                            from_table=target_table,
                            from_field=target_field,
                            from_pkey=target_pkey,
                            transform_dict=bool_to_int
                        )

                    if  j == 'measureDisabilities':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of disabled population
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )
                    
                    if  j == 'measureIncome':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of households below poverty
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )
                    
                    if  j == 'measureLimitedEnglishProficiency':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of population with limited english proficiency
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )
                    
                    if  j == 'measureVehicleAccess':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of households without vehicle access
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if  j == 'measureOlderPopulation':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of population over 75 years old
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if j == 'measureHispanicPopulation':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of Hispanic population 
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if  j == 'measureNonWhitePopulation':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of non white population 
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if j == 'measureYoungerPopulation':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate density of population under 18
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_equity_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if  j == 'measureVPI':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # calculate max of VPI
                        # delete previously created VPI field
                        arcpy.DeleteField_management(tmp_prioritization_layer, j)
                        arcpy.AddField_management(tmp_prioritization_layer, j, "SHORT")
                        target_table = os.path.join(standard_gdb, sub_params[i][j]['table'])
                        if not arcpy.Exists(target_table):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_table}'")
                            raise arcpy.ExecuteError
                        get_attr_from_polygons(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_table,
                            target_field=sub_params[i][j]['column'],
                            agg="max",
                            scratch_gdb=scratch_gdb
                        )

                    if j == 'measurePedestrianNetwork':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        search_dist_dict = {"Lines": "25 Feet", "Points": "25 Feet"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "SHORT")
                        target_layer = os.path.join(standard_gdb, sub_params["factorConnectivity"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        ped_type_col = sub_params["factorConnectivity"][j]["column"]
                        ped_network_filter = f"{ped_type_col} <> 'crosswalk_unmarked'"
                        tmp_ped_network = os.path.join(scratch_gdb, "tmp_ped_network")
                        copy_fc_with_fields(
                            in_fc=target_layer,
                            out_fc=tmp_ped_network,
                            fields=[ped_type_col],
                            scratch_gdb=scratch_gdb,
                            where_clause=ped_network_filter
                        )
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=tmp_ped_network,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )
                        arcpy.CalculateField_management(tmp_prioritization_layer, j, f"bool(!{j}! > 0)", "PYTHON3")

                    if  j == 'measureIntersectionDensity':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # density of intersections within .5 mile
                        arcpy.AddField_management(tmp_prioritization_layer, j, "DOUBLE")
                        target_layer = os.path.join(standard_gdb, sub_params["factorConnectivity"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist="0.5 miles"
                        )
                        buffer_area_acres = {}
                        cursor = arcpy.da.SearchCursor(tmp_half_mile_buffer, [primary_key, "SHAPE@AREA"])
                        for row in cursor:
                            buffer_area_acres[row[0]] = row[1]/43560 # area in acres
                        
                        cursor = arcpy.da.UpdateCursor(tmp_prioritization_layer, [primary_key, j])
                        for row in cursor:
                            row[1] = row[1]/buffer_area_acres[row[0]]
                            cursor.updateRow(row)
                        del cursor

                    if j == 'measurePopulationDensity':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # density population within .5 mile 
                        arcpy.AddField_management(tmp_prioritization_layer, j, "DOUBLE")
                        arcpy.AddField_management(tmp_half_mile_buffer, j, "DOUBLE")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        target_field = sub_params["factorPedTripPotential"][j]["column"]
                        calculate_density_by_intersect(
                            input_fc=tmp_half_mile_buffer,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            target_uid="p_id",
                            field_dict={j: target_field},
                            scratch_gdb=scratch_gdb,
                            area_multiplier=43560   # this is to get per acre value instead of per sq ft
                        )
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_half_mile_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if  j == 'measureEmploymentDensity':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # density of jobs within 1 mile 
                        arcpy.AddField_management(tmp_prioritization_layer, j, "DOUBLE")
                        arcpy.AddField_management(tmp_half_mile_buffer, j, "DOUBLE")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        target_field = sub_params["factorPedTripPotential"][j]["column"]
                        calculate_density_by_intersect(
                            input_fc=tmp_half_mile_buffer,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            target_uid="p_id",
                            field_dict={j: target_field},
                            scratch_gdb=scratch_gdb,
                            area_multiplier=43560   # this is to get per acre value instead of per sq ft
                        )
                        transfer_values(
                            to_table=tmp_prioritization_layer,
                            to_field=j,
                            to_pkey=primary_key,
                            from_table=tmp_half_mile_buffer,
                            from_field=j,
                            from_pkey=primary_key
                        )

                    if j == 'measureSchools':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        search_dist_dict = {"Lines": "0.5 miles", "Points": "0.5 miles"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "LONG")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )

                    if  j == 'measureHigherEducation':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        search_dist_dict = {"Lines": "1 miles", "Points": "1 miles"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "LONG")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )

                    if j == 'measureRetail':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        search_dist_dict = {"Lines": "0.25 miles", "Points": "0.25 miles"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "LONG")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )

                    if  j == 'measureMinorTransit':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # density of minor transit stops within .5 mile 
                        search_dist_dict = {"Lines": "0.125 miles", "Points": "0.25 miles"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "LONG")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )
  
                    if j == 'measureMajorTransit':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # density of major transit stops within .5 mile 
                        search_dist_dict = {"Lines": "0.25 miles", "Points": "0.5 miles"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "LONG")
                        target_layer = os.path.join(standard_gdb, sub_params["factorPedTripPotential"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )
  
                    if j == 'measureCrashes':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        search_dist_dict = {"Lines": "250 Feet", "Points": "500 Feet"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "FLOAT")
                        target_layer = os.path.join(standard_gdb, sub_params["factorSafety"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )
                        # convert crashes to crashes per mile for line features
                        if shp_type == "Lines":
                            arcpy.CalculateField_management(tmp_prioritization_layer, j, f"!{j}!/!shape.length@miles!")
  
                    if  j == 'measureSevereFatalCrashes':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        search_dist_dict = {"Lines": "250 Feet", "Points": "500 Feet"}
                        arcpy.AddField_management(tmp_prioritization_layer, j, "FLOAT")
                        target_layer = os.path.join(standard_gdb, sub_params["factorSafety"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=target_layer,
                            agg="sum",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type]
                        )
                        # convert crashes to crashes per mile for line features
                        if shp_type == "Lines":
                            arcpy.CalculateField_management(tmp_prioritization_layer, j, f"!{j}!/!shape.length@miles!")
  
                    if j == 'measureFunctionalClass':
                        arcpy.AddMessage('Generating calculations for: ' + sub_params[i][j]['name'])
                        # highest functional classification within 150 feet. Note on data: lower number functional classifications are assumed to be higher class.  
                        # filter some functional class values (criteria may need updating with new data)
                        target_layer = os.path.join(standard_gdb, sub_params["factorSafety"][j]["table"])
                        if not arcpy.Exists(target_layer):
                            arcpy.AddError(f"Data corresponding to '{j}' is not present in standard GDB. \n Expected dataset: '{target_layer}'")
                            raise arcpy.ExecuteError
                        fclass_col = sub_params["factorSafety"][j]["column"]
                        # exclude freeways and consider any fuctional 0 the same as 7
                        tmp_fclass_layer = os.path.join(scratch_gdb, "tmp_fclass_layer")
                        copy_fc_with_fields(
                            in_fc=target_layer,
                            out_fc=tmp_fclass_layer,
                            fields=[fclass_col],
                            scratch_gdb=scratch_gdb,
                            where_clause=f"{fclass_col} <> 1"
                        )
                        arcpy.CalculateField_management(tmp_fclass_layer, fclass_col, f"7 if !{fclass_col}! == 0 else !{fclass_col}!")
                        arcpy.AddField_management(tmp_prioritization_layer, j, "SHORT")
                        search_dist_dict = {"Lines": "150 Feet", "Points": "150 Feet"}
                        count_within_dist(
                            in_fc=tmp_prioritization_layer,
                            field=j,
                            input_uid=primary_key,
                            target_fc=tmp_fclass_layer,
                            agg="min",
                            scratch_gdb=scratch_gdb,
                            search_dist=search_dist_dict[shp_type],
                            target_field=fclass_col
                        )


        arcpy.AddMessage("Writing raw scores to excel file")
        ws = wb["Raw_Scores"]
        
        pkey_row = dict()
        i = 2
        cursor = arcpy.da.SearchCursor(tmp_prioritization_layer, [primary_key])
        for row in cursor:
            ws.cell(row=i, column=1).value = row[0]
            pkey_row[row[0]] = i
            i += 1
        
        # write other measures info to the excel file
        for i in sub_params.keys():
            for j, v in sub_params[i].items():
                if _get_by_name(parameters, j).value:
                    cursor = arcpy.da.SearchCursor(tmp_prioritization_layer, [primary_key, j])
                    for row in cursor:
                        ws.cell(row=pkey_row[row[0]], column=v['cell_index']).value = row[1]
        
        wb.save(output_excel)
        return 


def _get_by_name(param_list, name):
    list = [param_list[i].name for i in range(len(param_list))]
    return param_list[list.index(name)]