# -*- coding: utf-8 -*-
from xml.etree.ElementInclude import include
import arcpy
import os
import shutil
import pandas as pd
from openpyxl import load_workbook
from importlib import reload

import python.scripts_utilities.utils_functions
reload(python.scripts_utilities.utils_functions)
from python.scripts_utilities.utils_functions import * 

import python.scripts_utilities.importer_config
reload(python.scripts_utilities.importer_config)
from python.scripts_utilities.importer_config import *

import python.scripts_utilities.prioritization_config
reload(python.scripts_utilities.prioritization_config)
from python.scripts_utilities.prioritization_config import * 


class Prioritization_Join_Workbook(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Join Prioritization Workbook to Layer Tool"
        self.description = "Returns a feature class file with prioritization scores joined from workbook."
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""

        # 1 Features to prioritize
        prioritizion_features = arcpy.Parameter(
            name="prioritizion_features",
            displayName="Prioritization Feature Layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input"
        )    

        primary_key = arcpy.Parameter(
            name="primary_key",
            displayName="Primary Key",
            datatype="GPString",
            parameterType="Required",
            direction="Input"
        ) 
        primary_key.filter.type = "ValueList"
        primary_key.filter.list = ["Select a Layer of Prioritization Features"]

        gdb_dir = arcpy.Parameter(
            name="gdb_dir",
            displayName="Folder with Standardized and Scratch Geodatabases",
            datatype="DEFolder",
            parameterType="Required",
            enabled=True,
            direction="Input"
        )

        prioritization_workbook = arcpy.Parameter(
            name="prioritization_workbook",
            displayName="Excel Workbook with Prioritization Scores",
            datatype="DEFile",
            parameterType="Required",
            enabled=True,
            direction="Input"
        ) 
        prioritization_workbook.filter.list = ['xlsx']

        output_fc = arcpy.Parameter(
            name="output_fc",
            displayName="Output Feature Class with Scores",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Output"
        )
        output_fc.value = os.path.join(arcpy.env.workspace, "Prioritization_Scores_Joined")

        params = [
            prioritizion_features,
            primary_key,
            gdb_dir,
            prioritization_workbook,
            output_fc,
            ]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""


        if parameters[0].valueAsText != None and parameters[0].enabled: 
            fc_attributes =[f.name for f in arcpy.ListFields(parameters[0].valueAsText) if not f.required] 
            parameters[1].filter.list = fc_attributes
        
        # TODO consider commenting this out if we want the user to explicit give the location of output file
        # if parameters[2].value:
        #     parameters[4].enabled = True
        #     parameters[4].value = os.path.join(parameters[2].valueAstext,"Standardized_BMC_Ped_Infra_Assessment.gdb", "Prioritization_Results_Joined")

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        # list = [parameters[i].name for i in range(len(parameters))]
        prioritizion_features = _get_by_name(parameters, "prioritizion_features").valueAsText
        primary_key = _get_by_name(parameters, "primary_key").valueAsText
        gdb_dir = _get_by_name(parameters, "gdb_dir").valueAsText
        prioritization_workbook = _get_by_name(parameters, "prioritization_workbook").valueAsText
        output_fc = _get_by_name(parameters, "output_fc").valueAsText

        # # check that any factors selected have weights in template
        # input_wt_ws = template_wb[inputs_sheet]
        # for i in sub_params.keys():
        #     if _get_by_name(parameters, i).value and (input_wt_ws[f"{factor_wt_col}{factor_params[i]['wt_cell_row']}"].value > 0):

        # ensure the proper gdb exists in the location specified
        standard_gdb = os.path.join(gdb_dir, "Standardized_BMC_Ped_Infra_Assessment.gdb")
        scratch_gdb = os.path.join(gdb_dir, "Scratch_BMC_Ped_Infra_Assessment.gdb")

        # Set environment settings
        arcpy.env.workspace = gdb_dir
        workspaces = arcpy.ListWorkspaces("*", "FileGDB")
        arcpy.env.overwriteOutput = True

        if not bool(set(workspaces) == set([standard_gdb, scratch_gdb])):
            arcpy.AddError("Geodatabase names are not valid. This can result from A) incorrect geodatabase names, B) missing geodatabases, C) too many geodatabases in folder. Please refer to the manual for the required geodatabases.")
            raise arcpy.ExecuteError

        shp_info = arcpy.Describe(prioritizion_features)
        if shp_info.ShapeType in ["Point", "Multipoint"]:
            shp_type = "Points"
        elif shp_info.ShapeType in ["Polyline", "Multipolyline"]:
            shp_type = "Lines"
        else:
            arcpy.AddError('The shape type of the shapefile you uploaded is not one of the following formats: {0}'.format(["Polyline", "Multipolyline", "Point", "Multipoint"]))
            raise arcpy.ExecuteError
        
        # check if pkey field is unique
        # TODO check if this is necessary?
        if not is_field_unique(fc=prioritizion_features, field=primary_key):
            arcpy.AddError("The primary key field supplied is either missing value or is being used by more than one features.")
            raise arcpy.ExecuteError
        
        ## ensure excel template is set up right
        wb = load_workbook(prioritization_workbook, read_only=True, data_only=True)
        inputs_sheet = "Inputs - Weights"
        raw_scores_sheet = "Raw_Scores"
        weighted_vars_sheet = "WeightedScores - Variables"
        weighted_factors_sheet = "WeightedScores - Factors"
        ws_list = [inputs_sheet, raw_scores_sheet, weighted_vars_sheet, weighted_factors_sheet]

        if len([s for s in ws_list if s not in wb.sheetnames]) > 0:
            arcpy.AddError(f"Excel Template provided does not have a sheet with titles - '{[s for s in ws_list if s not in wb.sheetnames]}'.\n It has the following sheets {wb.sheetnames}")
            raise arcpy.ExecuteError
        
        ws = wb[inputs_sheet]
        include_dict = defaultdict(int)
        cell_index_dict = dict()
        # read which variables and factors are included in prioritization workbook
        for i in sub_params.keys():
            for j, v in sub_params[i].items():
                if ws[f"{variable_include_col}{v['cell_index']}"].value == True:
                    include_dict[i] = 1
                    include_dict[j] = 1
                    cell_index_dict[j] = v['cell_index']
                else:
                    include_dict[j] = 0
            if include_dict[i] == 0:
                pass
        
        # create a copy of the prioritization layer in the scratch database
        tmp_prioritization_layer = os.path.join(scratch_gdb, "tmp_prioritization_layer")
        copy_fc_with_fields(prioritizion_features, tmp_prioritization_layer, fields=[primary_key], scratch_gdb=scratch_gdb)

        # write raw scores to a table
        tmp_raw_scores = os.path.join(scratch_gdb, "tmp_raw_scores")
        arcpy.ExcelToTable_conversion(prioritization_workbook, tmp_raw_scores, raw_scores_sheet, 1, '')

        tmp_weighted_scores = os.path.join(scratch_gdb, "tmp_weighted_scores")
        arcpy.ExcelToTable_conversion(prioritization_workbook, tmp_weighted_scores, weighted_vars_sheet, 1, '')

        tmp_factor_scores = os.path.join(scratch_gdb, "tmp_factor_scores")
        arcpy.ExcelToTable_conversion(prioritization_workbook, tmp_factor_scores, weighted_factors_sheet, 1, '')
        # excel to table is not able to read the otehr tabl values for some reason
        # arcpy.TableToTable_conversion(f"{prioritization_workbook}\T_WeightedScores___Factors$_", scratch_gdb, "tmp_factor_scores")

        num_rows = int(arcpy.management.GetCount(tmp_raw_scores).getOutput(0))
        # arcpy.AddMessage(f"Num rows in raw scores = {num_rows}")

        # # read weighted scores sheet from excel
        # ws = wb[weighted_vars_sheet]
        # write raw and scaled variable scores
        arcpy.AddMessage("Writing raw and weighted-rescaled scores")
        for k, v in include_dict.items():
            if include_dict[k] == 1 and k[:7] == "measure":
                f_name = k[7:]
                arcpy.AddMessage(f"Joining variable score for '{f_name}'")
                # get field type for a given variable
                f_type = [f.type for f in arcpy.ListFields(tmp_raw_scores) if f.name == f_name][0]
                # arcpy.AddMessage(f"Field type of '{f_name}' = '{f_type}'")
                arcpy.AddField_management(tmp_prioritization_layer, f"raw_{f_name}", f_type)
                arcpy.AddField_management(tmp_prioritization_layer, f"variable_{f_name}", "DOUBLE")

                # read raw values
                raw_val_dict = dict()
                cursor = arcpy.da.SearchCursor(tmp_raw_scores, ["ID", f_name], where_clause = "ID IS NOT NULL")
                for row in cursor:
                    raw_val_dict[str(row[0])] = row[1]
                del cursor
                # arcpy.AddMessage(raw_val_dict)

                # read weighted scores
                weighted_scores_dict = dict()
                cursor = arcpy.da.SearchCursor(tmp_weighted_scores, ["ID", f_name], where_clause = "ID IS NOT NULL")
                for row in cursor:
                    weighted_scores_dict[str(row[0])] = str(row[1])
                del cursor
                # arcpy.AddMessage(weighted_scores_dict)

                cursor = arcpy.da.UpdateCursor(tmp_prioritization_layer, [primary_key, f"raw_{f_name}", f"variable_{f_name}"])
                for row in cursor:
                    # if str(row[0]) in raw_val_dict.keys():
                    row[1] = raw_val_dict[str(row[0])]
                    # if str(row[0]) in weighted_scores_dict.keys():
                    row[2] = float(weighted_scores_dict[str(row[0])])
                    cursor.updateRow(row)
                del cursor
        
        # factor scores
        arcpy.AddMessage("Writing factor scores")
        for k in include_dict.keys():
            if include_dict[k] == 1 and k[:6] == "factor":
                f_name = k[6:]
                arcpy.AddMessage(f"Joining factor score for '{f_name}'")
                arcpy.AddField_management(tmp_prioritization_layer, f"factor_{f_name}", "DOUBLE")
                factor_scores_dict = dict()
                cursor = arcpy.da.SearchCursor(tmp_factor_scores, ["ID", f_name], where_clause = "ID IS NOT NULL")
                for row in cursor:
                    factor_scores_dict[str(row[0])] = str(row[1])
                del cursor
                # arcpy.AddMessage(factor_scores_dict)

                cursor = arcpy.da.UpdateCursor(tmp_prioritization_layer, [primary_key, f"factor_{f_name}"])
                for row in cursor:
                    # if str(row[0]) in factor_scores_dict.keys():
                    row[1] = float(factor_scores_dict[str(row[0])])
                    cursor.updateRow(row)
                del cursor
        
        arcpy.AddField_management(tmp_prioritization_layer, f"Total_Score", "DOUBLE")
        total_scores_dict = dict()
        cursor = arcpy.da.SearchCursor(tmp_factor_scores, ["ID", "Total"], where_clause = "ID IS NOT NULL")
        for row in cursor:
            total_scores_dict[str(row[0])] = str(row[1])
        del cursor
        # arcpy.AddMessage(total_scores_dict)

        cursor = arcpy.da.UpdateCursor(tmp_prioritization_layer, [primary_key, "Total_Score"])
        for row in cursor:
            # if str(row[0]) in raw_val_dict.keys():
            row[1] = float(total_scores_dict[str(row[0])])
            cursor.updateRow(row)
        del cursor
        
        arcpy.management.CopyFeatures(tmp_prioritization_layer, output_fc)
        return True


def _get_by_name(param_list, name):
    list = [param_list[i].name for i in range(len(param_list))]
    return param_list[list.index(name)]